//
// Title:   CorrelationDialogImpl.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qmessagebox.h>
#include <qobject.h>
#include <qcombobox.h>
#include <qtextview.h>
#include <qlabel.h>
#include <qpixmap.h>
#include <qpainter.h>
#include "CorrelationDialogImpl.h"

/**
 * Constructor for the correlation dialog of the watermarking plugin.
 *
 * Dimension of the canvas is 470 x 298 and displays a plot containing a
 * threshold and a correlation value for each x-Coordinate. Only the first
 * 470 values will be plotted. The remaining values are simply ignored.
 * Threshold values are plotted using blue color, whereas the correlation
 * value is plotted green in case the correlation is larger than the 
 * threshold value or red if this is not the case.
 *
 * @param correlationList	List containing the <tt>float</tt> correlation values.
 * @param thresholdList		List containing the <tt>float</tt> threshold values.
 * @param parent			Standard QT dialog parameter.
 * @param name				Standard QT dialog parameter.
 * @param modal				Standart QT dialog parameter.
 * @param f					Standard QT dialog parameter.
 */
CorrelationDialogImpl::CorrelationDialogImpl(QList<float>* correlationList, 
											 QList<float>* thresholdList, 
											 QWidget* parent, 
											 const char* name, 
											 bool modal, 
											 WFlags f)
											 : CorrelationDialog(parent, name, modal, f) 
{

	// create the pixmap containing the graph and the corresponding paint object
	QPixmap pixmap(470,298);
	QPainter paint(&pixmap);

	// clear the background (white color)
	paint.setBackgroundColor(Qt::white);
	paint.eraseRect(0,0,470,298); 

	// contains the number of correlations, which are higher thatn the threshold
	int successCounter = 0;

	// reset the counter of the number of displayed correlation and thershold values
	int count = 0;

	// width of area for one data value (at least 1)
	int width = 470/correlationList->count();
	width = (width<1)?1:width;

	// calculate the difference (this will be visible as spaces on the left and right side)
	int difference = 470 - width*correlationList->count();
	count = difference / 2;

	// iterate over all the threshold values
	float* threshold = thresholdList->first();
	for (float* correlation = correlationList->first(); correlation!=0; correlation = correlationList->next()) {

		// the canvas can only display 470 correlation and threshold values
		// ignore the values which can no longer be displayed
		if (count>=470) {
			break;
		}

		// paint all the points for one data value
		// to get better visual diagrams we stretch the point graph
		for (int offset=0; offset<width; offset++) {
			// paint the threshold value using blue color
			paint.setPen(Qt::blue);
			paint.drawPoint(count+offset,288-(*threshold)*278);

			// paint the 0, 0.25, 0.5, 0.75 and 1.0 delimiter line every two pixels
			if ((count+offset)%2==1) {
				paint.setPen(Qt::gray);
				paint.drawPoint(count+offset,288-(0.00)*278+1);
				paint.drawPoint(count+offset,288-(0.25)*278+1);
				paint.drawPoint(count+offset,288-(0.50)*278+1);
				paint.drawPoint(count+offset,288-(0.75)*278+1);
				paint.drawPoint(count+offset,288-(1.00)*278+1);
			}

			// paint the watermark correlation value
			if (*correlation>*threshold) {
				paint.setPen(Qt::green);
				successCounter++;
			} else {
				paint.setPen(Qt::red);
			}
			paint.drawPoint(count+offset,288-(*correlation)*278);
		}
		
		// get next threshold value and increment the counter
		threshold = thresholdList->next();
		count+=width;
	}
	
	// add the pixmap containing the plot to the dialog box
	pixmapLabel->setPixmap(pixmap);

	// set the comment text
	if (successCounter>0) {
		commentLabel->setText("SUCCESS: The object contains your watermark!");
	} else {
		commentLabel->setText("FAILURE: Your watermark could not be found in the object!");
	}
} 

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
