//
// Title:   GenericVector.h
// Created: November 28, 2002
// Authors: Daniel Cotting
// 
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef GENERICVECTOR_H
#define GENERICVECTOR_H

#include "../../../Core/Interfaces/src/SurfelInterface.h"

////////////////////////////////////////////////////////////////////////////
//
// Generic vector
// This class inherits from Vector3D
// Note that not all methods have been overridden to reflect the changes.
// Refer to the source code for further information.
//

template <unsigned tNofValues>
class GenericVector: public Vector3D {

public:

	/////////////////////////////////////////////////////////////////
	// constructors
	/////////////////////////////////////////////////////////////////	

	GenericVector () {
		nofValues = tNofValues;
		// check if we are more than a simple 3D vector
		if (nofValues>3) {
			additionalValues = new float[nofValues-3];
		}
	}

	~GenericVector() {
		if (nofValues>3) {
			assert(additionalValues!=NULL);
			delete[] additionalValues;
			additionalValues=NULL;
		}
	}


	/////////////////////////////////////////////////////////////////

	inline GenericVector (float* pp)
	{ 
		nofValues = tNofValues;
		if (nofValues>3) {
			additionalValues = new float[nofValues-3];
		}
		p[0] = pp[0];
		p[1] = pp[1];
		p[2] = pp[2];
		for (uint i=3; i<nofValues; i++) {
				additionalValues[i-3] = pp[i];
		}
		
	};


	/////////////////////////////////////////////////////////////////
	// overloaded operators
	/////////////////////////////////////////////////////////////////
	
	inline float& operator[] (int Index)	
	{
		if (Index<3) {
			return (p[Index]);
		} else {
			return (additionalValues[Index-3]);
		}
	};


	/////////////////////////////////////////////////////////////////

	inline const float& operator[] (int Index) const	
	{
		if (index<3) {
			return (p[Index]);
		} else {
			return (additionalValues[Index-3]);
		}
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator= (const GenericVector& P)
	{
		p[0] = P[0];
		p[1] = P[1];
		p[2] = P[2];
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] = P[i];
		}
		return (*this);
	};

	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator+= (GenericVector P)
	{
		p[0] += P[0];
		p[1] += P[1];
		p[2] += P[2];
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] += P[i];
		}
		return (*this);
	};

	
	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator-= (GenericVector P)
	{
		p[0] -= P[0];
		p[1] -= P[1];
		p[2] -= P[2];
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] -= P[i];
		}
		return (*this);
	};
		

	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator*= (GenericVector P)
	{
		p[0] *= P[0];
		p[1] *= P[1];
		p[2] *= P[2];
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] *= P[i];
		}
		return (*this);
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator*= (float s)
	{
		p[0] *= s;
		p[1] *= s;
		p[2] *= s;
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] *= s;
		}
		return (*this);
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator/= (GenericVector P)
	{
		p[0] /= P[0];
		p[1] /= P[1];
		p[2] /= P[2];
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] /= P[i];
		}
		return (*this);
	};

	/////////////////////////////////////////////////////////////////

	inline GenericVector& operator/= (float s)
	{
		p[0] /= s;
		p[1] /= s;
		p[2] /= s;
		for (uint i=3; i<nofValues; i++) {
			additionalValues[i-3] /= s;
		}
		return (*this);
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector operator+ (const GenericVector P) const
	{
		GenericVector res;
		res[0] = p[0] + P[0];
		res[1] = p[1] + P[1];
		res[2] = p[2] + P[2];
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] + P[i];
		}
		return (res); 
	};

	
	/////////////////////////////////////////////////////////////////

	inline GenericVector operator- (const GenericVector P) const
	{
		GenericVector res;
		res[0] = p[0] + P[0];
		res[1] = p[1] + P[1];
		res[2] = p[2] + P[2];
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] + P[i];
		}
		return (res); 
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector operator- () const
	{
		GenericVector res;
		res[0] = -p[0];
		res[1] = -p[1];
		res[2] = -p[2];
		for (uint i=3; i<nofValues; i++) {
			res[i] = -additionalValues[i-3];
		}
		return (res); 
	};

	/////////////////////////////////////////////////////////////////

	inline GenericVector operator* (const GenericVector P) const
	{
		GenericVector res;
		res[0] = p[0] * P[0];
		res[1] = p[1] * P[1];
		res[2] = p[2] * P[2];
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] * P[i];
		}
		return (res); 
	};

	/////////////////////////////////////////////////////////////////

	inline GenericVector operator* (float s) const
	{
		GenericVector res;
		res[0] = p[0] * s;
		res[1] = p[1] * s;
		res[2] = p[2] * s;
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] * s;
		}
		return (res); 
	};


	/////////////////////////////////////////////////////////////////

	inline GenericVector operator/ (const GenericVector P) const
	{
		GenericVector res;
		res[0] = p[0] / P[0];
		res[1] = p[1] / P[1];
		res[2] = p[2] / P[2];
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] / P[i];
		}
		return (res); 
	};

	/////////////////////////////////////////////////////////////////

	inline GenericVector operator/ (float s) const
	{
		GenericVector res;
		res[0] = p[0] / s;
		res[1] = p[1] / s;
		res[2] = p[2] / s;
		for (uint i=3; i<nofValues; i++) {
			res[i] = additionalValues[i-3] / s;
		}
		return (res); 
	};

protected:
	float* additionalValues; 
	uint nofValues;
};

#endif  // __GENERICVECTOR_H_


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
