//
// Title:   PCA.cpp (originally PCA_v01.c)
// Created: November 28, 2002
// Authors: Daniel Cotting (for adaptation to PointShop3D, main authors see below)
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//
////////////////////////////////////////////////////////////////////////////////
// Based on code from the following project:
////////////////////////////////////////////////////////////////////////////////
// PCA Plug-In for BioArraySoftware Environment( BASE ) 
// homepage http://www.ngelab.org
// BASE - homepage http://base.thep.lu.se/
// Copyright (C) 2002 Ron Hart, Greg Voronin(rhgv)
//
// Write to: See Below
// This file is part of PCA_v01, a Principal Components Analysis Plug_In for BASE.
//
// PCA_v01 is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// PCA_v01 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//
// Creation of this software and documentation was supported by a grant from the 
// New Jersey Commission on Spinal Cord Research. Support is also acknowledged 
// from the Spinal Cord Injury Project of the W.M. Keck Center for Collaborative 
// Neuroscience of Rutgers University.
//	
//	Authors:
//
//	Gregory O. Voronin 	govkn@yahoo.com
//	Ronald P. Hart 		   rhart@andromeda.rutgers.edu
//	Neuroscience Gene Expression Laboratory( www.ngelab.org ),
//	W.M. Keck Center Center for Collaborative Neuroscience
//	Rutgers University
//	Piscataway, NJ 08854
//

/** 
 * @file PCA.cpp
 *
 * Principal Component Analysis code using GSL.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "Random.h"
#include "PCA.h"


/**
 * Main function for testing purpose of the PCA code.
 *
 * @return Always 0.
 */
int main( void ) {

	// define the size of the test matrix
	int matrixRows = 100; 
	int matrixCols = 100;

	// allocate the test matrix and the covariance matrix
	gsl_matrix *a = gsl_matrix_alloc(matrixRows, matrixCols);
	gsl_matrix *s = gsl_matrix_alloc(matrixCols, matrixCols);

	// initialize random number generator and randomize the test matrix
	Random random(3615);
	for (int currentRow=0; currentRow<matrixRows; currentRow++) {
		for (int currentCol=0; currentCol<matrixCols; currentCol++) {
			gsl_matrix_set( a, currentRow, currentCol, random.getUniform());
		}
	}

	// initialize PCA data structures
	PCA pca;
	PCA::EigenList eigenList = pca.initializeEigenList(matrixCols);

	// calculate the covariance matrix of the test matrix
	pca.createCovarianceMatrix(a, s);

	// calculate the eigen system of the covariance matrix
	pca.calculateEigenSystem(s, eigenList ); 
	
	// output the covariance matrix and the eigen system
	pca.outputCovarMatrix( eigenList, s );
	pca.outputEigenSystem( eigenList );

	gsl_matrix_free(a);
	gsl_matrix_free(s);
	
	return 0;
}


/**
 * Creates a list into which the eigenvalues and eigenvectors can be stored. 
 *
 * The implementation of the linked list for the eigensystem is a standard linked
 * list implementation available in numerous code cookbooks and data structures 
 * and algorithm's course text books. 
 * The annotation string's location does not imply a particular attachment to the
 * eigenvector or eigenvalue value in the linked list. It is a merely a convienent
 * place to store this information and use a later time for output. 
 *
 * @param The dimension of the matrix the eigenvalues have to be calculated for.
 * @return The list containing storage space for the eigenvalues and vectors.
 */
PCA::EigenList PCA::initializeEigenList(int col) {

	// allocate the head of the list
	EigenList eigenHead;
	eigenHead = (EigenList)malloc(sizeof(EigenRec));
	
	// if allocation failed, then print error message end exit.
	if (eigenHead == NULL) {
		printf("failure to allocate space for head of eigen system list\n");
		exit(1);
	}

	// add the following elements
	eigenHead->next = NULL;
	for (int i = 0; i < col; i++ ) {
		addEigenLink( "unknown", eigenHead );
	}

	// return the head of the list
	return eigenHead;

}

/**
 * Adds an element to the list.
 *
 * @param annotation The annotation of the element to add.
 * @param elist The head of the list to add a new element to.
 */
void PCA::addEigenLink( char *annotation, EigenList elist ) {

	// new element 
	EigenRec *newRec;
	
	// create head in case the head of the list does not exist
	if ( elist == NULL ) {
		elist = (EigenList)malloc(sizeof(EigenRec));
		printf("allocating a new list\n");
	}
	
	
	newRec = (EigenList)malloc(sizeof(EigenRec));

	if ( newRec == NULL ) {
		printf( "Falied to alloc space for an EigenRec.\n" );
		exit( 1 );
	}

	
	newRec->next = NULL;
	newRec->annotation = ( char *)malloc( strlen( annotation ) + 1 );
	strcpy( newRec->annotation, annotation );

	while( elist->next != NULL ) {
		elist = elist->next;
	}

	elist->next = newRec;
	
}


/* The following functions perform the PCA calculation. */

void PCA::createCovarianceMatrix( gsl_matrix *m, gsl_matrix *sm) {

	double columnSum, columnAvg, scaling;

	gsl_matrix *am;
	gsl_matrix *at;

	int maxRow = m->size1;
	int maxCol = m->size2;
	
	am = gsl_matrix_alloc( maxRow, maxCol );

	columnSum = 0;
	columnAvg = 0;
	for (int i = 0; i < maxCol; i++ ) { 
		for (int j = 0; j < maxRow; j++ ) {
			columnSum = columnSum + gsl_matrix_get( m, j, i );
		}
		columnAvg = columnSum / ( m->size1 );
		for (j = 0; j< maxRow; j++ ) {
			gsl_matrix_set( am, j, i, columnAvg );
		}
		columnSum = 0;
		columnAvg = 0;
	}

	/* the matrix m will now be turned into the difference matrix */
	gsl_matrix_sub( m, am );

	scaling = 1.0/(m->size1 - 1);	

	at = gsl_matrix_alloc( m->size2, m->size1 );
	gsl_matrix_transpose_memcpy( at, m );

	
	gsl_blas_dgemm( CblasNoTrans, CblasNoTrans, 1.0, at, m, 0.0, sm );
	gsl_matrix_scale( sm, scaling );

	/*
	printf( "The variance-covariance matrix:\n" );
	for (unsigned int ui = 0; ui < sm->size1; ui++ ) {
		for (unsigned int uj = 0; uj < sm->size2; uj++ ) {
			printf( " %f ", gsl_matrix_get( sm, ui, uj ) );
		}
		printf( "\n" );
	}
	printf( "\n" );
	*/

	/* the matrix s is now the variance-covariance matrix */

	// gsl_matrix_free( m );
	gsl_matrix_free( am );
	gsl_matrix_free( at );


}

void PCA::calculateEigenSystem( gsl_matrix *m, EigenList eList ) {

	unsigned int index;

	gsl_vector *eigenValues;
	gsl_matrix *working_m, *eigenVectors;	
	gsl_eigen_symmv_workspace *w;

	eigenValues = gsl_vector_alloc( m->size2 );

	working_m = gsl_matrix_alloc( m->size1, m->size2 );
	gsl_matrix_memcpy( working_m, m );	

	eigenVectors = gsl_matrix_alloc( m->size2, m->size2 );
	w = gsl_eigen_symmv_alloc( m->size2 );
	
	gsl_eigen_symmv( working_m, eigenValues, eigenVectors, w );
	gsl_eigen_symmv_free( w );
	
	gsl_eigen_symmv_sort( eigenValues, eigenVectors, GSL_EIGEN_SORT_VAL_DESC );

	/* assign the eigensystem list to the appropriate values */
	

	eList = eList->next;
	index = 0;
	while ( eList != NULL && index < eigenValues->size ) {

		eList->eigenvalue = gsl_vector_get( eigenValues, index );
		eList->cols = eigenValues->size;
		eList->eigenvector = gsl_vector_alloc( eigenValues->size );
		gsl_matrix_get_col( eList->eigenvector, eigenVectors, index );

		index = index + 1;
		eList = eList->next;

	}

	gsl_vector_free(eigenValues);
	gsl_matrix_free(working_m);
	gsl_matrix_free(eigenVectors);
}

/* The following functions create the HTML and text files displaying 
    the results of the PCA computation. */

void PCA::outputCovarMatrix( EigenList eList, gsl_matrix *m ) {

	unsigned int col, row;
	EigenList head = eList; /* maintain a pointer to the head of the list */
	FILE *fp1 = fopen( "Covariance.html", "w" );

	if ( fp1 != NULL ) {

		/* Begin of html output of the variance-covariance matrix */
		fprintf( fp1, "<html>" );

		fprintf( fp1, "<table border=\"3\" cellpadding=\"3\" cellspacing=\"1\">" );
		fprintf( fp1, "<caption><b>Variance-Covariance Matrix of the Data</b></caption>" );
		fprintf( fp1, "<tr><td></td>" );

		/* Place experiment names across the top of the table */
		
		/* Skip the head of the list */
		eList = eList->next; 
		while( eList != NULL ) {
		
		
			fprintf( fp1, "<td align=\"center\"><font color=\"blue\">%s</font></td>", eList->annotation );
			eList = eList->next;		
			
		}	
		
		fprintf( fp1, "</tr>" );
	
		/* Begin the rows and columns of the variance-covariance matrix by returning to the head of the list
                   to obtain experiment names and used the variance-covariance matrix to display the results           */

		
		
		eList = head;
		eList = eList->next;
		row = 0;
		while ( eList != NULL ) {	
		
			fprintf( fp1, "</tr>" );
			fprintf( fp1, "<td align=\"center\"><font color=\"blue\">%s</font></td>", eList->annotation );

			for (col = 0; col<m->size2; col++ ) {

				fprintf( fp1, "<td align=\"center\">%f</td>", gsl_matrix_get( m, row, col ) );

			} 

			eList = eList->next;
			row = row + 1;
			fprintf( fp1, "<tr>" );

		}
				
		/* End of the html output of the variance-covariance matrix */		
		fprintf( fp1, "</table>" );
		fprintf( fp1, "</html>" );
		fclose( fp1 );

	}

}

void PCA::outputEigenSystem( EigenList eList ) {
	
	int i, index;
	

	EigenList head;

	FILE *fp1 = fopen( "Eigensystem.html", "w" );

	head = eList;

	if ( fp1 != NULL ) {

		/* The files have both been created and can be written too */

		/* write to the HTML file */
		fprintf( fp1, "<html><head>" );
		fprintf( fp1, "<body><Br></Br>" );
		
		eList = eList->next; /* Skip the head of the list */
		fprintf( fp1, "<table border=\"3\" cellpadding=\"3\" cellspacing=\"1\">" );
		fprintf( fp1, "<caption><b>Principal Components Table</b></caption>" );
			

		fprintf( fp1, "<tr>" );

		fprintf( fp1, "<td align=\"center\"><font color=\"red\">Eigenvalue</font></td>" );
	
		fprintf( fp1, "</tr>" );
			
		eList = head;
		eList = eList->next;
		index = 1;
		while ( eList != NULL ) {

			fprintf( fp1, "<tr>" );
			fprintf( fp1, "<td align=\"center\"><font color=\"red\">%f</font></td>", eList->eigenvalue ); 
			fprintf( fp1, "<td align=\"center\"><font color=\"blue\">principal component %d</td>", index );

			for ( i = 0; i<eList->cols; i++ ) {

				fprintf( fp1, "<td align=\"center\">%f</font></td>", gsl_vector_get( eList->eigenvector, i ) );

			}

			fprintf( fp1, "</tr>" );

			eList = eList->next;
			index = index + 1;
		}	

		fprintf( fp1, "</table>" );		
	
		
		fclose( fp1 );

	}

}

