//
// Title:   PCAPatchBuilder.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file PCAPatchBuilder.cpp
 *
 * Takes a list of surfels and creates patches each containing a part of the surfels.
 * This is used to create smaller collections of surfels that can be processed
 * independently from each other. This results in better robustness against cropping
 * attacks. Furthermore some algorithms work faster on patches than on an entire object.
 * This implementation uses PCA based recursive space subdivision.
 */

#include <qobject.h>
#include <stdlib.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "MySurfelList.h"
#include "Patch.h"
#include "PCA.h"
#include "PCAPatchBuilder.h"
#include "WatermarkInterface.h"

/** 
 * Constructor of the class <tt>PCAPatchBuilder</tt>.
 *
 * Random number generator is only used to get some good results without having to 
 * process all the surfels (it has no security relevance).That's why we set its seed
 * to a deterministic value in this constructor.
 * 
 * @param watermark The watermark to be embedded later on. The bits of the watermark
 *					are not accessed. Only the size of the watermark is needed here.
 */
PCAPatchBuilder::PCAPatchBuilder(WatermarkInterface* watermark): random(Random(3615)) {
	nofMinimumPointsPerPatch = watermark->getSizeInBits();
}

/**
 * Destructor of the class <tt>PCAPatchBuilder</tt>.
 */
PCAPatchBuilder::~PCAPatchBuilder() {
}

/**
 * Takes a list of srufels and creates parches each containing a part of the surfels.
 * 
 * @param surfelList	The list of surfels that have to be subdivided into patches.
 * @return				The list of patches that have been generated.
 */
PatchBuilderInterface::PatchList* PCAPatchBuilder::buildPatches(MySurfelList* surfelList ) {

	// patch data structure
	PatchList* patchList = new PatchList();

	// start with all points and partition them recursively
	this->recursivePartition(patchList, surfelList);

	// return result
	return patchList;
}

/** 
 * Recursively partitions the surfels in <tt>surfelList</tt> and adds the patches
 * to <tt>patchList</tt>.
 *
 * @param patchList The list to add the patches to.
 * @param surfelList This list of surfels that still have to be partitioned.
 */
void PCAPatchBuilder::recursivePartition(PatchBuilderInterface::PatchList* patchList, MySurfelList* surfelList) {
	
	// we only have to subpartition if there are enough points to create another subdivision
	if (surfelList->count()>=2*nofMinimumPointsPerPatch) {

		// allocate matrices to compute PCA over 100 random points of the list
		unsigned int matrixRows = 100; 
		unsigned int matrixCols = 3;

		gsl_matrix *a = gsl_matrix_alloc(matrixRows, matrixCols);
		gsl_matrix *s = gsl_matrix_alloc(matrixCols, matrixCols);

		// calculate the probability that one specific surfel has to be included in the list
		unsigned int sizeOfMySurfelList = surfelList->count();
		float probability = 100.0f/sizeOfMySurfelList;

		// init data structure for max extent
		float minX = FLT_MAX;
		float minY = FLT_MAX;
		float minZ = FLT_MAX;
		float maxX = FLT_MIN;
		float maxY = FLT_MIN;
		float maxZ = FLT_MIN;

		// create the initial matrix and at the same time compute the center of mass
		unsigned int surfelNumber = 0;
		double centerX = 0.0;
		double centerY = 0.0;
		double centerZ = 0.0;

		// add 100 random points to the matrix
		while (surfelNumber < 100) {
			for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
				if (random.getUniform()<probability) {
					Vector3D position = surfel->getPosition();
					gsl_matrix_set( a, surfelNumber, 0, (double)position[0]);
					gsl_matrix_set( a, surfelNumber, 1, (double)position[1]);
					gsl_matrix_set( a, surfelNumber, 2, (double)position[2]);
					centerX += position[0];
					centerY += position[1];
					centerZ += position[2];
					minX = (position[0]<minX ? position[0]:minX);
					minY = (position[1]<minY ? position[1]:minY);
					minZ = (position[2]<minZ ? position[2]:minZ);
					maxX = (position[0]>maxX ? position[0]:maxX);
					maxY = (position[1]>maxY ? position[1]:maxY);
					maxZ = (position[2]>maxZ ? position[2]:maxZ);
					surfelNumber++;
					if (surfelNumber==100) break;
				}
			}
		}

		// calculate final approximation of the center of mass
		centerX /= surfelNumber;
		centerY /= surfelNumber;
		centerZ /= surfelNumber;

		// calculate final approximation of the maximum extent of the surfel collection
		float maxExtent = (maxX-minX);
		maxExtent = (maxY-minY>maxExtent ? maxY-minY:maxExtent);
		maxExtent = (maxZ-minZ>maxExtent ? maxZ-minZ:maxExtent);

		// compute the PCA
		PCA pca;
		pca.createCovarianceMatrix(a, s);
		PCA::EigenList eigenList = pca.initializeEigenList(matrixCols);
		PCA::EigenList eigenHead = eigenList;
		pca.calculateEigenSystem(s, eigenList ); 

		// Extract the largest eigenvector
		eigenList = eigenList->next;
		float eX = gsl_vector_get(eigenList->eigenvector, 0);
		float eY = gsl_vector_get(eigenList->eigenvector, 1);
		float eZ = gsl_vector_get(eigenList->eigenvector, 2);

		// free eigenlist
		while (eigenList!=NULL) {
			PCA::EigenList nextElement = eigenList->next;
			free(eigenList->annotation);
			gsl_vector_free(eigenList->eigenvector);
			free(eigenList);
			eigenList = nextElement;
		}
		free(eigenHead);

		// free matrices 
		gsl_matrix_free(a);
		gsl_matrix_free(s);

		// calculate coefficients of plane equation through the center of mass with the eigenvector as the normal
		float planeA = eX;
		float planeB = eY;
		float planeC = eZ;
		float planeD = -(planeA*centerX+planeB*centerY+planeC*centerZ);
	
		// process all points, split them and add them to the sublists
		MySurfelList subList1;
		MySurfelList subList2;
		float modificationFactor = maxExtent/100;	// to get two sets containing both enough points the plane has to be moved sometimes
		int modificationDirection = 0;				// direction in which we last moved the separation plane

		// loop while we have not found a good separation plane
		while (true) {

			// clear both sublists
			subList1.clear();
			subList2.clear();

			// subdivide all the surfels based on their position with respect to the plane
			for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
				Vector3D position = surfel->getPosition();
				if (planeA*(float)(position[0])+planeB*(float)(position[1])+planeC*(float)(position[2])+planeD>0) {
					subList1.append(surfel);
				} else {
					subList2.append(surfel);
				}
			}

			// one of the subsets does not contain enough points?
			if (subList2.count()<nofMinimumPointsPerPatch) {
				// planeD verringern
				planeD -= modificationFactor;
				if (modificationDirection==1) {
					// we've changed in the other direction previously and the change was too large
					// so let's reduce the modification factor.
					modificationFactor /= 10;
				}
				modificationDirection = -1;
			} else if (subList1.count()<nofMinimumPointsPerPatch) {
				// planeD vergroessern
				planeD += modificationFactor;
				if (modificationDirection==-1) {
					// we've changed in the other direction previously and the change was too large
					// so let's reduce the modification factor.
					modificationFactor /= 10;
				}
				modificationDirection = 1;
			} else {

				// we have found a good plane with 2 good subsets.
				break;
			}
		}

		// recursively partition the two subsets
		recursivePartition(patchList, &subList1);
		recursivePartition(patchList, &subList2);

	} else {

		// we do not have enough surfels to subdivide the list again, so let's 
		// put all the remaining surfels into a patch.
		// don't subpartition, but create a patch
		Patch* patch = new Patch();
		for (SurfelInterface* surfel = surfelList->first(); surfel!=0; surfel=surfelList->next()) {
			patch->append(surfel);
		}

		// add patch to the patch list
		patchList->append(patch);

		// output user information
		// qDebug("Points in patch: %d\r\n", surfelList->count());
	}
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* PCAPatchBuilder::getModuleName() const {
	return "PCA Patch Builder";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PCAPatchBuilder::getModuleDescription() const {
	return "Builds patches by recursively splitting the remaining space using a plane perpendicular to the principal component calculated with the help of PCA";
}


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
