//
// Title:   PCAPatchBuilder.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PCAPATCHBUILDER_H_
#define __PCAPATCHBUILDER_H_

#include <qobject.h>

#include "../../../Core/Interfaces/src/PluginInterface.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "MySurfelList.h"
#include "PatchBuilderInterface.h"
#include "WatermarkInterface.h"
#include "Random.h"

/**
 * Takes a list of surfels and creates patches each containing a part of the surfels.
 * This is used to create smaller collections of surfels that can be processed
 * independently from each other. This results in better robustness against cropping
 * attacks. Furthermore some algorithms work faster on patches than on an entire object.
 * This implementation uses PCA based recursive space subdivision.
 */
class PCAPatchBuilder : public PatchBuilderInterface {

public:
	
	/** 
	 * Constructor of the class <tt>PCAPatchBuilder</tt>.
	 * 
	 * @param watermark The watermark to be embedded later on. The bits of the watermark
	 *					are not accessed. Only the size of the watermark is needed here.
	 */
	PCAPatchBuilder(WatermarkInterface* watermark);

	/**
	 * Destructor of the class <tt>PCAPatchBuilder</tt>.
	 */
	~PCAPatchBuilder();

	/**
	 * Takes a list of srufels and creates parches each containing a part of the surfels.
	 * 
	 * @param surfelList	The list of surfels that have to be subdivided into patches.
	 * @return				The list of patches that have been generated.
	 */
	PatchList* buildPatches(MySurfelList* surfelList);

	/**
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	const char* getModuleName() const;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	const char* getModuleDescription() const;

private:

	/** 
	 * Recursively partitions the surfels in <tt>surfelList</tt> and adds the patches
	 * to <tt>patchList</tt>.
	 *
	 * @param patchList The list to add the patches to.
	 * @param surfelList This list of surfels that still have to be partitioned.
	 */
	void recursivePartition(PatchList* patchList, MySurfelList* surfelList);

	/** 
	 * Random number generator.
	 */
	Random random;
};

#endif  //__PCAPATCHBUILDER_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
