//
// Title:   PlaneProjection.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/** 
 * @file PlaneProjection.cpp
 *
 * Projects a patch of the original object onto the surface of the verification object.
 */

#include <math.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "MySurfelList.h"
#include "PlaneProjection.h"
#include "Patch.h"

/**
 * Projects the patch <tt>originalSurfelPatch</tt> onto the surface of the verification 
 * object defined by <tt>verificationSurfelList</tt>.
 *
 * @param originalSurfelPatch The patch to project.
 * @param verificationSurfelList The surface to project the patch onto.
 * @return The projected patch.
 */
SurfelCollection* PlaneProjection::projectPatch(
	Patch* originalSurfelPatch, 
	MySurfelList* verificationSurfelList) 
{
	// the following code projects each surfel of the original object onto a plane defined
	// by the three nearest neighbours of the verification surface.

	int i;
	SurfelInterface* surfel;

	// create data structures for the neighbourhood search on the verification surface
	int nofVerificationSurfels = verificationSurfelList->count();
	Vector3D* verificationPositions = new Vector3D[nofVerificationSurfels];
	for (i = 0, surfel = verificationSurfelList->first(); i < nofVerificationSurfels; i++, surfel = verificationSurfelList->next()) {
		verificationPositions[i] = surfel->getPosition();
	}

	// initialize the neighbourhood search objects
	NeighbourHood* neighbourHoodOnVerificationSurface = new NeighbourHood(verificationPositions, nofVerificationSurfels);

	// create data structure which will hold the result
	SurfelCollection* projectedPatch = new SurfelCollection();
	SurfelInterface* projectedSurfel;

	// iterate over all the surfels of the original object and project them onto the modified surface
	int nofOriginalSurfels = originalSurfelPatch->count();
	for (i = 0, surfel = originalSurfelPatch->first(); i < nofOriginalSurfels; i++, surfel = originalSurfelPatch->next()) {

		// set the source point to the position of the surfel to project.
		Vector3D surfelPosition = surfel->getPosition();
		neighbourHoodOnVerificationSurface->setSourcePoint(surfelPosition);
		// get the nearest three neighbours
		neighbourHoodOnVerificationSurface->setNofNeighbours(3);

		// we will get three positions and calculate the plane equation
		Vector3D point1 = neighbourHoodOnVerificationSurface->getNeighbourPosition(0);
		Vector3D point2 = neighbourHoodOnVerificationSurface->getNeighbourPosition(1);
		Vector3D point3 = neighbourHoodOnVerificationSurface->getNeighbourPosition(2);
		Vector3D normal = Vector3D::crossProduct(point2-point1, point3-point1);
		float constD = -(normal[0]*point1[0]+normal[1]*point1[1]+normal[2]*point1[2]);

		// the plane is now defined by the normal and the constant D
		// we will now project the original surfel onto that surface
		float lambda = (-constD-normal[0]*surfelPosition[0]-normal[1]*surfelPosition[1]-normal[2]*surfelPosition[2]);
		lambda /= (normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
		Vector3D projectedPosition = surfelPosition + lambda*normal;

		//qDebug("%f %f %f / %f %f %f / %f %f %f", 
		//	point1[0],point1[1],point1[2],point2[0],point2[1],point2[2],point3[0],point3[1],point3[2]);

		//qDebug("%f %f %f -> %f %f %f", 
		//	surfelPosition[0],surfelPosition[1],surfelPosition[2],projectedPosition[0],projectedPosition[1],projectedPosition[2]);

		// create new surfel object
		// add the projected point to the patch
		projectedSurfel = projectedPatch->addSurfel(surfel, false);

		// modify the Position. this can be accomplished by using the scene even though 
		// the surfel is not really displayed
		Scene::getInstance()->setSurfelPosition(projectedSurfel, projectedPosition, false);
	}

	// clean up
	delete neighbourHoodOnVerificationSurface;
	delete verificationPositions;

	// return the result
	return projectedPatch;
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* PlaneProjection::getModuleName() const {
	return "Plane Projection";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PlaneProjection::getModuleDescription() const {
	return "Projects points of one patch onto the other surface by computing a plane going through the three nearest neighbours of the point to project.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
