//
// Title:   SteganoWatermarking.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __STEGANOWATERMARKING_H_
#define __STEGANOWATERMARKING_H_

#include <qobject.h>
#include "Random.h"
#include "Patch.h"
#include "MySurfelList.h"
#include "WatermarkInterface.h"
#include "WatermarkingInterface.h"
#include "PatchBuilderInterface.h"

/**
 * Implements a steganography scheme, which allows to embed a text in a 3D object.
 */
class SteganoWatermarking: public WatermarkingInterface {

public:

	/**
	 * Constructor of the class <tt>SteganoWatermarking</tt>.
	 */
	SteganoWatermarking():random(new Random(3615)){};

	/**
	 * Embeds a Stegano into the currently selected surfels or the active object.
	 * Although a selection is supported in the embedding process, the extraction procedure
	 * only supports PointShop3D objects and no selection. Therefore selections have to be converted 
	 * to objects before extraction.
	 *
	 * @param passPhrase The secret, user defined passphrase to use for the Stegano creation.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	void embedWatermark(const QString& passPhrase, WatermarkInterface* watermark);

	/**
	 * Extracts the private watermark by using the test object and the original object. Note that
	 * both geometries have to be available as objects. No selection is supported during extraction.
	 * That's the place where our private watermarking algorithm is significantly different from the 
	 * public one (here we need two objects: the original one and the object to be verified).
	 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
	 *
	 * @param passPhrase The passphrase to be used for extraction.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	void extractWatermark(const QString& passPhrase, WatermarkInterface* watermark);

	/**
	 * Returns the name of this module.
	 *
	 * @return "Steganography"
	 */
	const char* getModuleName() const;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	const char* getModuleDescription() const;

protected:

	/**
	 * Embeds the watermark into the patch.
	 *
	 * @param patch The patch to embed the watermark into.
	 * @param watermark The watermark to embed into the patch.
	 */
	void embedWatermarkIntoPatch(
		Patch* patch, 
		WatermarkInterface* watermark);

	/**
	 * Extracts watermark from a patch and a verification object.
	 *
	 * @param originalPatch				The patch to be processed (on the original object).
	 * @param verificationSurfelList	The list containing all the points from the object to be verified.
	 * @param originalWatermark			The watermark that is supposed to result from the extraction. It will be
	 *									compared with the watermark that is extracted and a correlation value will
	 *									be calculated.
	 * @return							The correlation value between the specified watermark and the extracted one.
	 */
	float extractWatermarkFromPatch(
		Patch* originalPatch,
		MySurfelList* verificationSurfelList,
		WatermarkInterface* originalWatermark);

	/**
	 * Random number generator.
	 */
	Random* random;
};

#endif __STEGANOWATERMARKING_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
