//
// Title:   WatermarkDialogImpl.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qobject.h>
#include <qcombobox.h>
#include <qtextview.h>
#include "WatermarkDialogImpl.h"
#include "WatermarkInterface.h"
#include "MajorityWatermark.h"

/**
 * Constructor for the dialog of the watermarking plugin.
 *
 * @param algorithms					Array of algorithms to display in the dialog box. 
 * @param parent						Standard QT dialog parameter.
 * @param name							Standard QT dialog parameter.
 * @param modal							Standart QT dialog parameter.
 * @param f								Standard QT dialog parameter.
 */
WatermarkDialogImpl::WatermarkDialogImpl
	(WatermarkingInterface** algorithms, QWidget* parent, const char* name, bool modal, WFlags f):
	WatermarkDialog(parent, name, modal, f), passphrase(*(new QString("")))
{
	// store the algorithms for later usage
	watermarkingAlgorithms = algorithms;

	// iterate over all the algorithms and add them to the user interface 
	// so that the user can select one of them
	uint index = 0;
	while (watermarkingAlgorithms[index]!=NULL) {

		// insert the watermarking name into a combobox
		ComboBox1->insertItem(watermarkingAlgorithms[index]->getModuleName());

		// we initially select the first watermarking algorithm
		if (index==0) {
			// add the description of the first algorithm to the description box of the UI
			TextView1->setText(watermarkingAlgorithms[index]->getModuleDescription());
			// mark the first algorithm as selected algorithm
			selectedWatermarkingAlgorithm = watermarkingAlgorithms[index];
		}

		// go to the next algorithm
		index++;
	}

	ComboBox2->insertItem("Geometry");
	ComboBox2->insertItem("Color");
	colorWatermark = false;

	ComboBox7->insertItem("Fast");
	ComboBox7->insertItem("Accurate");
	useFastVersion = true;

	// sets the default passphrase
	setPassphrase(QString(""));

	// set default watermark length
	watermarkBitLength = 56;

	// set default  watermark strength
	watermarkStrength = 0.001f;

	// set the default chip rate
	watermarkChipRate = 3;

	// sets the default frequency usage
	watermarkUsageDivider = 10;
} 
 
/**
 * Embeds the watermark by calling the appropriate function of the currently selected algorithm.
 * A watermark configured with the parameters the user has specified is passed to the watermarking
 * procedure.
 */
void WatermarkDialogImpl::embedWatermark()
{
	WatermarkInterface* watermark = new MajorityWatermark(watermarkBitLength,
		watermarkStrength, watermarkChipRate, watermarkUsageDivider, colorWatermark, useFastVersion);
	selectedWatermarkingAlgorithm->embedWatermark(passphrase, watermark);
}
  
/** 
 * Extracts the watermark by calling the appropriate function of the currently selected algorithm.
 * A watermark configured with the parameters the user has specified is passed to the watermarking
 * procedure.
 */
void WatermarkDialogImpl::extractWatermark()
{
	WatermarkInterface* watermark = new MajorityWatermark(watermarkBitLength,
		watermarkStrength, watermarkChipRate, watermarkUsageDivider, colorWatermark, useFastVersion);
	selectedWatermarkingAlgorithm->extractWatermark(passphrase, watermark);
} 

/**
 * Sets the passphrase to the specified value.
 *
 * @param passphraseIn The new passphrase to use for the watermark.
 */
void WatermarkDialogImpl::setPassphrase(const QString& passphraseIn)
{
	passphrase = passphraseIn;  
}  

/**
 * Sets the watermark bit length to the specified value.
 *
 * @param watermarkBitLengthIn The new watermark length value to use (in bits).
 */
void WatermarkDialogImpl::setWatermarkBitLength(int watermarkBitLengthIn)
{
	watermarkBitLength = watermarkBitLengthIn;  
}  

/**
 * Sets the watermark chip rate to the specified value.
 *
 * @param watermarkChipRateIn The new watermark chip rate to use (number of repetitions).
 */
void WatermarkDialogImpl::setWatermarkChipRate(int watermarkChipRateIn)
{
	watermarkChipRate = watermarkChipRateIn;  
}  

/**
 * Sets the watermark usage divider to the specified value.
 *
 * @param watermarkUsageDividerIn The new watermark usage divider to use (fraction of frequencies to use for embedding).
 */
void WatermarkDialogImpl::setWatermarkUsageDivider(int watermarkUsageDividerIn)
{
	watermarkUsageDivider = watermarkUsageDividerIn;  
}  

/** 
 * Sets the strength (alpha value) of the watermark.
 *
 * @param watermarkStrength The strength of the watermark in Promille.
 */
void WatermarkDialogImpl::setWatermarkStrengthInPromille(int watermarkStrengthInPromille)
{
	watermarkStrength = ((float)watermarkStrengthInPromille)/1000.0f;  
}  

/**
 * Sets the selected algorithm to the algorithm having the same name as the passed parameter.
 *
 * @param algorithmIn The name of the algorithm to select.
 */
void WatermarkDialogImpl::setAlgorithm(const QString& algorithmIn)
{
	// check all algorithms to get a match
	uint index = 0;
	while (watermarkingAlgorithms[index]!=NULL) {

		// does the algorithm at the current index position match the name?
		if (strcmp(watermarkingAlgorithms[index]->getModuleName(), algorithmIn)==0) {

			// we have found the corresponding algorithm 
			// so update the description text and set the new selected algorithm
			TextView1->setText(watermarkingAlgorithms[index]->getModuleDescription());
			selectedWatermarkingAlgorithm = watermarkingAlgorithms[index];
		}

		// go to the next index
		index++;
	}
}  

/**
 * Sets whether the watermark has to be applied to color or position.
 *
 * @param colorWatermarkIn Defines whether the watermark has to be applied to the color attribute.
 */
void WatermarkDialogImpl::setIsColorWatermark(int colorWatermarkIn)
{
	colorWatermark = (colorWatermarkIn==1);
}  

/**
 * Sets whether the fast version of the embedding is to be used.
 *
 * @param selectionIndex Index of the user selection (0 for fast, 1 for accurate).
 */
void WatermarkDialogImpl::setUseFastVersion(int selectionIndex)
{
	useFastVersion = (selectionIndex==0);
}  

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
